from django.conf import settings
from django.http import Http404
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import ensure_csrf_cookie
from graphene_file_upload.django import FileUploadGraphQLView

SHOULD_EXPOSE_GRAPHIQL = settings.DEBUG or getattr(
    settings, "BIFROST_EXPOSE_GRAPHIQL", False
)


class BifrostView(View):
    playground = SHOULD_EXPOSE_GRAPHIQL

    @method_decorator(ensure_csrf_cookie)
    def get(self, request):
        if self.playground:
            return render(
                request,
                "bifrost/playground.html",
                {"url": request.build_absolute_uri()},
            )

        raise Http404()

    @method_decorator(ensure_csrf_cookie)
    def post(self, request):
        return FileUploadGraphQLView.as_view(graphiql=False)(request)



# Generated by Django 3.1.5 on 2021-01-11 08:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("bifrost", "0002_stubmodel")]

    operations = [migrations.DeleteModel(name="StubModel")]



import inspect
from types import MethodType

import graphene
from wagtail.core.models import Page

from .registry import registry
from .types.streamfield import StreamFieldInterface

streamfield_types = []


def register_streamfield_block(cls):
    base_block = None
    for block_class in inspect.getmro(cls):
        if block_class in registry.streamfield_blocks:
            base_block = registry.streamfield_blocks[block_class]

    streamfield_types.append(
        {
            "cls": cls,
            "type_prefix": "",
            "interface": StreamFieldInterface,
            "base_type": base_block,
        }
    )

    return cls


def register_graphql_schema(Query=None, Mutation=None, Subscription=None):
    if Query:
        registry.queries.append(Query)
    if Mutation:
        registry.mutations.append(Mutation)
    if Subscription:
        registry.subscriptions.append(Subscription)


def register_query_field(field_name, plural_field_name=None, query_params=None):
    from .types.structures import QuerySetList
    from .utils import resolve_queryset

    if not plural_field_name:
        plural_field_name = field_name + "s"

    def inner(cls):
        field_type = lambda: registry.models[cls]
        field_query_params = query_params or {"id": graphene.Int()}

        def Mixin():
            # Generic methods to get all and query one model instance.
            def resolve_singular(self, _, info, **kwargs):
                # If no filters then return nothing,
                if not kwargs:
                    return None

                try:
                    # If is a Page then only query live/public pages.
                    if issubclass(cls, Page):
                        return cls.objects.live().public().get(**kwargs)

                    return cls.objects.get(**kwargs)
                except:
                    return None

            def resolve_plural(self, _, info, **kwargs):
                qs = cls.objects
                if issubclass(cls, Page):
                    qs = cls.objects.live().public().order_by("-first_published_at")

                # Filter queryset for field_query_params
                # Important: each field_query_param must be indexed in the model:
                # index.FilterField("field_query_param")
                qs = qs.filter(
                    **{
                        key: kwargs.get(key)
                        for (key, value) in field_query_params.items()
                        if key not in ["id", "token"] and key in kwargs
                    }
                )

                return resolve_queryset(qs.all(), info, **kwargs)

            # Create schema and add resolve methods
            schema = type(cls.__name__ + "Query", (), {})
            setattr(
                schema, field_name, graphene.Field(field_type, **field_query_params)
            )
            setattr(schema, plural_field_name, QuerySetList(field_type))
            setattr(
                schema, "resolve_" + field_name, MethodType(resolve_singular, schema)
            )
            setattr(
                schema,
                "resolve_" + plural_field_name,
                MethodType(resolve_plural, schema),
            )
            return schema

        # Send schema to Bifrost schema.
        register_graphql_schema(Query=Mixin())
        return cls

    return inner


def register_paginated_query_field(
    field_name,
    plural_field_name=None,
    query_params=None,
    required=False,
    plural_required=False,
    plural_item_required=False,
):
    from .types.structures import PaginatedQuerySet
    from .utils import resolve_paginated_queryset

    if not plural_field_name:
        plural_field_name = field_name + "s"

    def inner(cls):
        field_type = lambda: registry.models[cls]
        field_query_params = query_params

        field_query_params = query_params or {"id": graphene.Int()}

        if issubclass(cls, Page):
            field_query_params["slug"] = graphene.Argument(
                graphene.String, description=ugettext_lazy("The page slug.")
            )
            field_query_params["token"] = graphene.Argument(
                graphene.String, description=ugettext_lazy("The preview token.")
            )

        def Mixin():
            # Generic methods to get all and query one model instance.
            # @test(cls)
            def resolve_singular(self, _, info, **kwargs):
                if hasattr(cls, "bifrost_queryset"):
                    qs = cls.bifrost_queryset(info, **kwargs)
                else:
                    qs = cls.objects

                # If no filters then return nothing.
                if not kwargs:
                    return None

                try:
                    # If is a Page then only query live/public pages.
                    if issubclass(cls, Page):
                        if "token" in kwargs and hasattr(
                            cls, "get_page_from_preview_token"
                        ):
                            return cls.get_page_from_preview_token(kwargs["token"])

                        return qs.live().public().get(**kwargs)

                    return qs.get(**kwargs)
                except:
                    return None

            def resolve_plural(self, _, info, **kwargs):
                if hasattr(cls, "bifrost_queryset"):
                    qs = cls.bifrost_queryset(info, **kwargs)
                else:
                    qs = cls.objects

                if issubclass(cls, Page):
                    qs = qs.live().public().order_by("-first_published_at")

                # Filter queryset for field_query_params
                # Important: each field_query_param must be indexed in the model:
                # index.FilterField("field_query_param")
                qs = qs.filter(
                    **{
                        key: kwargs.get(key)
                        for (key, value) in field_query_params.items()
                        if key not in ["id", "token"] and key in kwargs
                    }
                )

                return resolve_paginated_queryset(qs.all(), info, **kwargs)

            # Create schema and add resolve methods
            schema = type(cls.__name__ + "Query", (), {})

            singular_field_type = field_type
            if required:
                singular_field_type = graphene.NonNull(field_type)

            setattr(
                schema,
                field_name,
                graphene.Field(singular_field_type, **field_query_params),
            )

            plural_field_type = field_type
            if plural_item_required:
                plural_field_type = graphene.NonNull(field_type)

            setattr(
                schema,
                plural_field_name,
                PaginatedQuerySet(
                    plural_field_type,
                    cls,
                    required=plural_required,
                    **field_query_params
                ),
            )

            setattr(
                schema, "resolve_" + field_name, MethodType(resolve_singular, schema)
            )
            setattr(
                schema,
                "resolve_" + plural_field_name,
                MethodType(resolve_plural, schema),
            )
            return schema

        # Send schema to Bifrost schema.
        register_graphql_schema(Query=Mixin())
        return cls

    return inner



# from .core import (
#     Page,
#     Site,
#     User,
#     # mixins
#     InfoQueryMixin,
#     PagesQueryMixin,
# )
# from .auth import (
#     LoginMutation,
#     LogoutMutation,
#     # mixins
#     AuthQueryMixin,
# )
# from .documents import (
#     Document,
#     # mixins
#     DocumentQueryMixin,
# )
# from .forms import FormField, FormError
# from .images import (
#     Image,
#     # mixins
#     ImageQueryMixin,
# )
# from .settings import Settings, SettingsQueryMixin
# from .snippets import SnippetsQueryMixin
# # noinspection PyUnresolvedReferences
# from . import converters    # noqa: F401

# __all__ = [
#     # core
#     'Page',
#     'Settings',
#     'Site',
#     'User',
#     # auth
#     'AuthQueryMixin',
#     'LoginMutation',
#     'LogoutMutation',
#     # documents
#     'Document',
#     # forms
#     'FormError',
#     'FormField',
#     # images
#     'Image',
#     # mixins
#     'DocumentQueryMixin',
#     'ImageQueryMixin',
#     'InfoQueryMixin',
#     'MenusQueryMixin',
#     'PagesQueryMixin',
#     'SettingsQueryMixin',
#     'SnippetsQueryMixin',
# ]

# # menus

# try:
#     from django.conf import settings
#     if 'wagtailmenus' not in settings.INSTALLED_APPS:
#         raise ImportError()  # pragma: no cover

#     from .menus import MenusQueryMixin, Menu, MenuItem, SecondaryMenu, SecondaryMenuItem  # noqa: F401

#     __all__.extend([
#         # menus
#         'Menu',
#         'MenuItem',
#         'SecondaryMenu',
#         'SecondaryMenuItem',
#     ])

#     HAS_WAGTAILMENUS = True
# except ImportError:  # pragma: no cover
#     def MenusQueryMixin():
#         # type: ignore
#         class _MenusQueryMixin:
#             pass
#         return _MenusQueryMixin
#     HAS_WAGTAILMENUS = False



from django.apps import AppConfig


class Bifrost(AppConfig):
    name = "bifrost"

    def ready(self):
        """
        Import all the django apps defined in django settings then process each model
        in these apps and create graphql node types from them. Then the schema file
        of all apps are imported.
        """
        from .actions import import_app_schema, import_apps, load_type_fields
        from .types.streamfield import register_streamfield_blocks

        import_apps()
        load_type_fields()
        register_streamfield_blocks()
        import_app_schema()



from django.conf.urls import url
from django.urls import path
from django.views.decorators.csrf import csrf_exempt
from graphene_subscriptions.consumers import GraphqlSubscriptionConsumer

from .views import BifrostView

# Traditional URL routing
urlpatterns = [url(r"^graphql", csrf_exempt(BifrostView.as_view()))]

websocket_urlpatterns = [path("graphql/", GraphqlSubscriptionConsumer)]

# if SHOULD_EXPOSE_GRAPHIQL:
#     urlpatterns.append(url(r"^graphiql", graphiql))



import graphene
import graphql_jwt
from django.conf import settings

# django
from django.utils.text import camel_case_to_spaces
from graphql.validation.rules import NoUnusedFragments, specified_rules

# HACK: Remove NoUnusedFragments validator
# Due to the way previews work on the frontend, we need to pass all
# fragments into the query even if they're not used.
# This would usually cause a validation error. There doesn't appear
# to be a nice way to disable this validator so we monkey-patch it instead.


# We need to update specified_rules in-place so the change appears
# everywhere it's been imported

specified_rules[:] = [rule for rule in specified_rules if rule is not NoUnusedFragments]


def create_schema():
    """
    Root schema object that graphene is pointed at.
    It inherits its queries from each of the specific type mixins.
    """
    from .jwtauth.schema import ObtainJSONWebToken, ObtainPrivilegedJSONWebToken
    from .registry import registry
    from .types.documents import DocumentsQuery
    from .types.images import ImagesQuery
    from .types.pages import PagesQuery, PagesSubscription
    from .types.redirects import RedirectsQuery
    from .types.search import SearchQuery
    from .types.settings import SettingsQuery
    from .types.snippets import SnippetsQuery

    class Query(
        PagesQuery(),
        ImagesQuery(),
        DocumentsQuery(),
        SnippetsQuery(),
        SettingsQuery(),
        SearchQuery(),
        RedirectsQuery,
        *registry.queries,
        graphene.ObjectType,
    ):
        pass

    class Subscription(
        PagesSubscription(), *registry.subscriptions, graphene.ObjectType
    ):
        pass

    def mutation_parameters() -> dict:
        dict_params = {
            "token_auth": ObtainJSONWebToken.Field(),
            "privileged_token_auth": ObtainPrivilegedJSONWebToken.Field(),
            "verify_token": graphql_jwt.Verify.Field(),
            "refresh_token": graphql_jwt.Refresh.Field(),
            "revoke_token": graphql_jwt.Revoke.Field(),
        }

        dict_params.update(
            (camel_case_to_spaces(n).replace(" ", "_"), mut.Field())
            for n, mut in registry.forms.items()
        )
        return dict_params

    Mutations = type("Mutation", (graphene.ObjectType,), mutation_parameters())

    class Mutation(Mutations, *registry.mutations):
        pass

    return graphene.Schema(
        query=Query,
        mutation=Mutation,
        subscription=Subscription,
        types=list(registry.models.values()),
        auto_camelcase=getattr(settings, "BIFROST_AUTO_CAMELCASE", True),
    )


schema = create_schema()



# django
# time
from datetime import timedelta

from django.conf import settings

# graphql
from graphql import ResolveInfo

# settings
if hasattr(settings, "GRAPHQL_API"):
    SETTINGS = settings.GRAPHQL_API
else:  # pragma: no cover
    SETTINGS = {}

URL_PREFIX = SETTINGS.get("URL_PREFIX", {})
LOAD_GENERIC_SCALARS = SETTINGS.get("GENERIC_SCALARS", True)
RELAY = SETTINGS.get("RELAY", False)

GRAPHENE = {
    "SCHEMA": "bifrost.schema.schema",
    "MIDDLEWARE": ["graphql_jwt.middleware.JSONWebTokenMiddleware"],
}
GRAPHQL_JWT = {
    "JWT_ALLOW_ARGUMENT": True,
    "JWT_VERIFY_EXPIRATION": True,
    "JWT_LONG_RUNNING_REFRESH_TOKEN": True,
    "JWT_EXPIRATION_DELTA": timedelta(minutes=5),
    "JWT_REFRESH_EXPIRATION_DELTA": timedelta(days=7),
}

AUTHENTICATION_BACKENDS = [
    "graphql_jwt.backends.JSONWebTokenBackend",
    "django.contrib.auth.backends.ModelBackend",
]

# wagtail settings
try:
    from wagtail.contrib.settings.registry import registry as settings_registry
except ImportError:  # pragma: no cover
    settings_registry = None


def url_prefix_for_site(info: ResolveInfo):
    hostname = info.context.site.hostname
    return URL_PREFIX.get(hostname, info.context.site.root_page.url_path.rstrip("/"))



import graphene
import graphql_jwt
from django.contrib.auth import get_user_model
from graphene.types.generic import GenericScalar
from graphene_django import DjangoObjectType
from graphql import GraphQLError


# Create your registration related graphql schemes here.


class UserType(DjangoObjectType):
    class Meta:
        model = get_user_model()
        fields = ("id", "username")


class ObtainJSONWebToken(graphql_jwt.JSONWebTokenMutation):

    user = graphene.Field(UserType)

    @classmethod
    def resolve(cls, root, info, **kwargs):
        user = info.context.user

        if user.is_superuser:
            raise GraphQLError("Something went wrong")

        return cls(user=info.context.user)


class ObtainPrivilegedJSONWebToken(graphene.Mutation):

    token = graphene.String()
    payload = GenericScalar(required=True)
    refresh_expires_in = graphene.Int(required=True)

    class Arguments:
        token = graphene.String(required=True)

    @classmethod
    def Field(cls, *args, **kwargs):
        cls._meta.arguments.update(
            {
                get_user_model().USERNAME_FIELD: graphene.String(required=True),
                "password": graphene.String(required=True),
            }
        )
        return super().Field(*args, **kwargs)

    @classmethod
    @graphql_jwt.decorators.staff_member_required
    @graphql_jwt.decorators.token_auth
    def mutate(cls, root, info, **kwargs):
        return ObtainPrivilegedJSONWebToken()

    @classmethod
    def resolve(cls, root, info, **kwargs):
        cls()



"""
ASGI config for z1socialideas project.
It exposes the ASGI callable as a module-level variable named ``application``.
For more information on this file, see
https://docs.djangoproject.com/en/3.0/howto/deployment/asgi/
"""

import django

django.setup()

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter

from .urls import websocket_urlpatterns

application = ProtocolTypeRouter(
    {
        # (http->django views is added by default)
        "websocket": AuthMiddlewareStack(URLRouter(websocket_urlpatterns))
    }
)



from django.conf import settings
from graphene_django import DjangoObjectType

# graphql_jwt
from graphql_jwt.decorators import login_required
from wagtailmedia.models import Media


class MediaObjectType(DjangoObjectType):
    class Meta:
        """Can change over time."""

        model = Media
        exclude = ("tags",)

    @login_required
    def resolve_file(self, info, **kwargs):
        if self.file.url[0] == "/":
            return settings.BASE_URL + self.file.url
        return self.file.url



# graphene
import graphene

# graphene_django
from graphene_django.converter import Boolean, String


class FormField(graphene.ObjectType):
    name = graphene.Field(String)
    field_type = graphene.Field(String)
    help_text = graphene.Field(String)
    required = graphene.Field(Boolean)
    choices = graphene.Field(String)
    default_value = graphene.Field(String)
    label = graphene.Field(String)


class FormError(graphene.ObjectType):
    name = graphene.Field(String)
    errors = graphene.List(String)



import inspect

import graphene
import wagtail
import wagtail.documents.blocks
import wagtail.embeds.blocks
import wagtail.images.blocks
import wagtail.snippets.blocks
from django.conf import settings
from django.template.loader import render_to_string
from graphene.types import Scalar
from graphene_django.converter import convert_django_field
from wagtail.core import blocks
from wagtail.core.fields import StreamField
from wagtail.core.rich_text import expand_db_html
from wagtail.embeds.blocks import EmbedValue
from wagtail.embeds.embeds import get_embed
from wagtail.embeds.exceptions import EmbedException

from ..registry import registry


class GenericStreamFieldInterface(Scalar):
    @staticmethod
    def serialize(stream_value):
        return stream_value.stream_data


@convert_django_field.register(StreamField)
def convert_stream_field(field, registry=None):
    return GenericStreamFieldInterface(
        description=field.help_text, required=not field.null
    )


class StreamFieldInterface(graphene.Interface):
    id = graphene.String()
    block_type = graphene.String(required=True)
    field = graphene.String(required=True)
    raw_value = graphene.String(required=True)

    @classmethod
    def resolve_type(cls, instance, info):
        """
        If block has a custom Graphene Node type in registry then use it,
        otherwise use generic block type.
        """
        if hasattr(instance, "block"):
            mdl = type(instance.block)
            if mdl in registry.streamfield_blocks:
                return registry.streamfield_blocks[mdl]

            for block_class in inspect.getmro(mdl):
                if block_class in registry.streamfield_blocks:
                    return registry.streamfield_blocks[block_class]

        return registry.streamfield_blocks["generic-block"]

    def resolve_id(self, info, **kwargs):
        return self.id

    def resolve_block_type(self, info, **kwargs):
        return type(self.block).__name__

    def resolve_field(self, info, **kwargs):
        return self.block.name

    def resolve_raw_value(self, info, **kwargs):
        if isinstance(self, blocks.StructValue):
            # This is the value for a nested StructBlock defined via GraphQLStreamfield
            return serialize_struct_obj(self)
        if isinstance(self.value, dict):
            return serialize_struct_obj(self.value)

        return self.value


def generate_streamfield_union(graphql_types):
    class StreamfieldUnion(graphene.Union):
        class Meta:
            types = graphql_types

        @classmethod
        def resolve_type(cls, instance, info):
            """
            If block has a custom Graphene Node type in registry then use it,
            otherwise use generic block type.
            """
            mdl = type(instance.block)
            if mdl in registry.streamfield_blocks:
                return registry.streamfield_blocks[mdl]

            return registry.streamfield_blocks["generic-block"]

    return StreamfieldUnion


class StructBlockItem:
    id = None
    block = None
    value = None

    def __init__(self, id, block, value=""):
        self.id = id
        self.block = block
        self.value = value


def serialize_struct_obj(obj):
    rtn_obj = {}

    if hasattr(obj, "stream_data"):
        rtn_obj = []
        for field in obj.stream_data:
            rtn_obj.append(serialize_struct_obj(field["value"]))
    else:
        for field in obj:
            value = obj[field]
            if hasattr(value, "stream_data"):
                rtn_obj[field] = list(
                    map(
                        lambda data: serialize_struct_obj(data["value"]),
                        value.stream_data,
                    )
                )
            elif hasattr(value, "value"):
                rtn_obj[field] = value.value
            elif hasattr(value, "src"):
                rtn_obj[field] = value.src
            elif hasattr(value, "file"):
                rtn_obj[field] = value.file.url
            else:
                rtn_obj[field] = value

    return rtn_obj


class StructBlock(graphene.ObjectType):
    class Meta:
        interfaces = (StreamFieldInterface,)

    # blocks = graphene.List(graphene.NonNull(StreamFieldInterface), required=True)

    # def resolve_blocks(self, info, **kwargs):
    #     stream_blocks = []

    #     if issubclass(type(self.value), wagtail.core.blocks.stream_block.StreamValue):
    #         # self: StreamChild, block: StreamBlock, value: StreamValue
    #         stream_data = self.value.stream_data
    #         child_blocks = self.value.stream_block.child_blocks
    #     else:
    #         # This occurs when StreamBlock is child of StructBlock
    #         # self: StructBlockItem, block: StreamBlock, value: list
    #         stream_data = self.value
    #         child_blocks = self.block.child_blocks

    #     for field, value in stream_data.items():
    #         block = dict(child_blocks)[field]
    #         if issubclass(
    #             type(block), wagtail.core.blocks.ChooserBlock
    #         ) or not issubclass(type(block), blocks.StructBlock):
    #             if isinstance(value, int):
    #                 value = block.to_python(value)

    #         stream_blocks.append(StructBlockItem(field, block, value))

    #     return stream_blocks


class StreamBlock(StructBlock):
    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_blocks(self, info, **kwargs):
        stream_blocks = []
        for stream in self.value.stream_data:
            if type(stream) == tuple:
                # As of Wagtail 2.11 stream_data is a list of dicts (when lazy) or tuples
                # when not lazy. The tuple is (block_type, value, id) where value has been run through bulk_to_python()
                # @see https://github.com/wagtail/wagtail/pull/5976
                block_type, value, _ = stream
                block = self.value.stream_block.child_blocks[block_type]
            else:
                block_type = stream["type"]
                value = stream["value"]
                block = self.value.stream_block.child_blocks[block_type]
                if issubclass(
                    type(block), wagtail.core.blocks.ChooserBlock
                ) or not issubclass(type(block), blocks.StructBlock):
                    value = block.to_python(value)

            stream_blocks.append(StructBlockItem(block_type, block, value))
        return stream_blocks


class StreamFieldBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class CharBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class TextBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class EmailBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class IntegerBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class FloatBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class DecimalBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class RegexBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class URLBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class BooleanBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class DateBlock(graphene.ObjectType):
    value = graphene.String(format=graphene.String(), required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_value(self, info, **kwargs):
        format = kwargs.get("format")
        if format:
            return self.value.strftime(format)
        return self.value


class DateTimeBlock(DateBlock):
    class Meta:
        interfaces = (StreamFieldInterface,)


class TimeBlock(DateBlock):
    class Meta:
        interfaces = (StreamFieldInterface,)


class RichTextBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_value(self, info, **kwargs):
        # Allow custom markup for RichText
        return render_to_string(
            "wagtailcore/shared/richtext.html",
            {"html": expand_db_html(self.value.source)},
        )


class RawHTMLBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class BlockQuoteBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class ChoiceOption(graphene.ObjectType):
    key = graphene.String(required=True)
    value = graphene.String(required=True)


class ChoiceBlock(graphene.ObjectType):
    value = graphene.String(required=True)
    choices = graphene.List(graphene.NonNull(ChoiceOption), required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_choices(self, info, **kwargs):
        choices = []
        for key, value in self.block._constructor_kwargs["choices"]:
            choice = ChoiceOption(key, value)
            choices.append(choice)
        return choices


def get_media_url(url):
    if url[0] == "/":
        return settings.BASE_URL + url
    return url


def get_embed_url(instance):
    return instance.value.url if hasattr(instance, "value") else instance.url


def get_embed_object(instance):
    try:
        return get_embed(get_embed_url(instance))
    except EmbedException:
        pass


class EmbedBlock(graphene.ObjectType):
    value = graphene.String(required=True)
    url = graphene.String(required=True)
    embed = graphene.String()
    raw_embed = graphene.JSONString()

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_url(self, info, **kwargs):
        return get_media_url(get_embed_url(self))

    def resolve_raw_value(self, info, **kwargs):
        if isinstance(self, EmbedValue):
            return self
        return StreamFieldInterface.resolve_raw_value(info, **kwargs)

    def resolve_embed(self, info, **kwargs):
        embed = get_embed_object(self)
        if embed:
            return embed.html

    def resolve_raw_embed(self, info, **kwargs):
        embed = get_embed_object(self)
        if embed:
            return {
                "title": embed.title,
                "type": embed.type,
                "thumbnail_url": embed.thumbnail_url,
                "width": embed.width,
                "height": embed.height,
                "html": embed.html,
            }


class StaticBlock(graphene.ObjectType):
    value = graphene.String(required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)


class ListBlock(graphene.ObjectType):
    items = graphene.List(graphene.NonNull(StreamFieldInterface), required=True)

    class Meta:
        interfaces = (StreamFieldInterface,)

    def resolve_items(self, info, **kwargs):
        # Get the nested StreamBlock type
        block_type = self.block.child_block
        # Return a list of GraphQL types from the list of values
        return [StructBlockItem(self.id, block_type, item) for item in self.value]


registry.streamfield_blocks.update(
    {
        "generic-block": StreamFieldBlock,
        # blocks.CharBlock: CharBlock,
        # blocks.TextBlock: TextBlock,
        # blocks.EmailBlock: EmailBlock,
        # blocks.IntegerBlock: IntegerBlock,
        # blocks.FloatBlock: FloatBlock,
        # blocks.DecimalBlock: DecimalBlock,
        # blocks.RegexBlock: RegexBlock,
        # blocks.URLBlock: URLBlock,
        # blocks.BooleanBlock: BooleanBlock,
        # blocks.DateBlock: DateBlock,
        # blocks.TimeBlock: TimeBlock,
        # blocks.DateTimeBlock: DateTimeBlock,
        # blocks.RichTextBlock: RichTextBlock,
        # blocks.RawHTMLBlock: RawHTMLBlock,
        # blocks.BlockQuoteBlock: BlockQuoteBlock,
        # blocks.ChoiceBlock: ChoiceBlock,
        blocks.StreamBlock: StreamBlock,
        blocks.StructBlock: StructBlock,
        # blocks.StaticBlock: StaticBlock,
        # blocks.ListBlock: ListBlock,
        wagtail.embeds.blocks.EmbedBlock: EmbedBlock,
    }
)


def register_streamfield_blocks():
    from .documents import get_document_type
    from .images import get_image_type
    from .pages import PageInterface

    class PageChooserBlock(graphene.ObjectType):
        page = graphene.Field(PageInterface, required=True)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_page(self, info, **kwargs):
            return self.value.specific

    class DocumentChooserBlock(graphene.ObjectType):
        document = graphene.Field(get_document_type(), required=True)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_document(self, info, **kwargs):
            return self.value

    class ImageChooserBlock(graphene.ObjectType):
        image = graphene.Field(get_image_type(), required=True)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_image(self, info, **kwargs):
            return self.value

    class SnippetChooserBlock(graphene.ObjectType):
        snippet = graphene.String(required=True)

        class Meta:
            interfaces = (StreamFieldInterface,)

        def resolve_snippet(self, info, **kwargs):
            return self.value

    registry.streamfield_blocks.update(
        {
            # blocks.PageChooserBlock: PageChooserBlock,
            # wagtail.documents.blocks.DocumentChooserBlock: DocumentChooserBlock,
            # wagtail.images.blocks.ImageChooserBlock: ImageChooserBlock,
            # wagtail.snippets.blocks.SnippetChooserBlock: SnippetChooserBlock,
        }
    )



# python
from typing import List

# graphene_django
import graphene
from graphene_django import DjangoObjectType

# graphql
from graphql import ResolveInfo

# wagtailmenus
from wagtailmenus.models import FlatMenu, FlatMenuItem, MainMenu, MainMenuItem


class MenuItem(DjangoObjectType):
    class Meta:
        """Can change over time."""

        model = MainMenuItem


class Menu(DjangoObjectType):
    class Meta:
        """Can change over time."""

        model = MainMenu
        only_fields = ["max_levels", "menu_items"]


class SecondaryMenuItem(DjangoObjectType):
    class Meta:
        """Can change over time."""

        model = FlatMenuItem


class SecondaryMenu(DjangoObjectType):
    class Meta:
        """Can change over time."""

        model = FlatMenu
        only_fields = ["title", "handle", "heading", "max_levels", "menu_items"]


def MenusQueryMixin():
    class Mixin:
        main_menu = graphene.List(Menu)
        secondary_menu = graphene.Field(
            SecondaryMenu, handle=graphene.String(required=True)
        )
        secondary_menus = graphene.List(SecondaryMenu)

        def resolve_main_menu(self, _info: ResolveInfo) -> List[MainMenu]:
            return MainMenu.objects.all()

        def resolve_secondary_menus(self, _info: ResolveInfo) -> List[FlatMenu]:
            return FlatMenu.objects.all()

        def resolve_secondary_menu(self, _info, handle: ResolveInfo) -> FlatMenu:
            return FlatMenu.objects.filter(handle=handle).first()

    return Mixin
